# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long, consider-using-f-string, no-else-return, duplicate-string-formatting-argument, expression-not-assigned, too-many-locals, logging-fstring-interpolation, arguments-differ, abstract-method, logging-format-interpolation, broad-except


from tempfile import NamedTemporaryFile
from urllib.parse import urlparse
import subprocess
import requests

from azure.cli.core.azclierror import (
    RequiredArgumentMissingError,
    ValidationError,
    InvalidArgumentValueError,
    MutuallyExclusiveArgumentError,
    CLIError,
)
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.command_modules.appservice._create_util import (
    check_resource_group_exists,
)
from azure.cli.command_modules.acr.custom import acr_show
from azure.cli.core.commands.client_factory import get_mgmt_service_client
from azure.mgmt.containerregistry import ContainerRegistryManagementClient
from knack.log import get_logger

from msrestazure.tools import parse_resource_id, is_valid_resource_id, resource_id

from ._clients import ManagedEnvironmentClient, ContainerAppClient, GitHubActionClient, ContainerAppsJobClient

from ._utils import (
    get_randomized_name,
    get_profile_username,
    create_resource_group,
    get_resource_group,
    queue_acr_build,
    _get_acr_cred,
    create_new_acr,
    _get_default_containerapps_location,
    safe_get,
    is_int,
    create_service_principal_for_github_action,
    repo_url_to_name,
    get_container_app_if_exists,
    get_containerapps_job_if_exists,
    _ensure_location_allowed,
    register_provider_if_needed,
    validate_environment_location,
    list_environment_locations,
    format_location,
    is_docker_running,
    get_pack_exec_path,
    get_latest_buildpack_run_tag

)

from ._constants import (MAXIMUM_SECRET_LENGTH,
                         LOG_ANALYTICS_RP,
                         CONTAINER_APPS_RP,
                         ACR_IMAGE_SUFFIX,
                         ACR_TASK_TEMPLATE,
                         DEFAULT_PORT)

from .custom import (
    create_managed_environment,
    create_containerappsjob,
    containerapp_up_logic,
    list_containerapp,
    list_managed_environments,
    create_or_update_github_action,
)

from ._github_oauth import load_github_token_from_cache, get_github_access_token

logger = get_logger(__name__)


class ResourceGroup:
    def __init__(self, cmd, name: str, location: str, exists: bool = None):
        self.cmd = cmd
        self.name = name
        self.location = _get_default_containerapps_location(cmd, location)
        if self.location.lower() == "northcentralusstage":
            self.location = "eastus"
        self.exists = exists

        self.check_exists()

    def create(self):
        g = create_resource_group(self.cmd, self.name, self.location)
        self.exists = True
        return g

    def _get(self):
        return get_resource_group(self.cmd, self.name)

    def get(self):
        r = None
        try:
            r = self._get()
        except:  # pylint: disable=bare-except
            pass
        return r

    def check_exists(self) -> bool:
        if self.name is None:
            self.exists = False
        else:
            self.exists = check_resource_group_exists(self.cmd, self.name)
        return self.exists

    def create_if_needed(self):
        if not self.check_exists():
            if not self.name:
                self.name = get_randomized_name(get_profile_username())
            logger.warning(f"Creating resource group '{self.name}'")
            self.create()
        else:
            logger.warning(f"Using resource group '{self.name}'")  # TODO use .info()


class Resource:
    def __init__(
            self, cmd, name: str, resource_group: "ResourceGroup", exists: bool = None
    ):
        self.cmd = cmd
        self.name = name
        self.resource_group = resource_group
        self.exists = exists

        self.check_exists()

    def create(self, *args, **kwargs):
        raise NotImplementedError()

    def _get(self):
        raise NotImplementedError()

    def get(self):
        r = None
        try:
            r = self._get()
        except:  # pylint: disable=bare-except
            pass
        return r

    def check_exists(self):
        if self.name is None or self.resource_group.name is None:
            self.exists = False
        else:
            self.exists = self.get() is not None
        return self.exists


class ContainerAppEnvironment(Resource):
    def __init__(
            self,
            cmd,
            name: str,
            resource_group: "ResourceGroup",
            exists: bool = None,
            location=None,
            logs_key=None,
            logs_customer_id=None,
    ):

        super().__init__(cmd, name, resource_group, exists)
        if is_valid_resource_id(name):
            self.name = parse_resource_id(name)["name"]
            if "resource_group" in parse_resource_id(name):
                rg = parse_resource_id(name)["resource_group"]
                if resource_group.name != rg:
                    self.resource_group = ResourceGroup(cmd, rg, location)
        self.location = location
        self.logs_key = logs_key
        self.logs_customer_id = logs_customer_id

    def set_name(self, name_or_rid):
        if is_valid_resource_id(name_or_rid):
            self.name = parse_resource_id(name_or_rid)["name"]
            if "resource_group" in parse_resource_id(name_or_rid):
                rg = parse_resource_id(name_or_rid)["resource_group"]
                if self.resource_group.name != rg:
                    self.resource_group = ResourceGroup(
                        self.cmd,
                        rg,
                        self.location,
                    )
        else:
            self.name = name_or_rid

    def _get(self):
        return ManagedEnvironmentClient.show(
            self.cmd, self.resource_group.name, self.name
        )

    def create_if_needed(self, app_name):
        if not self.check_exists():
            if self.name is None:
                self.name = "{}-env".format(app_name).replace("_", "-")
            logger.warning(
                f"Creating {type(self).__name__} '{self.name}' in resource group {self.resource_group.name}"
            )
            self.create()
        else:
            logger.warning(
                f"Using {type(self).__name__} '{self.name}' in resource group {self.resource_group.name}"
            )  # TODO use .info()

    def create(self):
        register_provider_if_needed(self.cmd, LOG_ANALYTICS_RP)

        if self.location:
            self.location = validate_environment_location(self.cmd, self.location)

            env = create_managed_environment(
                self.cmd,
                self.name,
                location=self.location,
                resource_group_name=self.resource_group.name,
                logs_key=self.logs_key,
                logs_customer_id=self.logs_customer_id,
                disable_warnings=True,
            )
            self.exists = True

            return env
        else:
            res_locations = list_environment_locations(self.cmd)
            for loc in res_locations:
                try:
                    env = create_managed_environment(
                        self.cmd,
                        self.name,
                        location=loc,
                        resource_group_name=self.resource_group.name,
                        logs_key=self.logs_key,
                        logs_customer_id=self.logs_customer_id,
                        disable_warnings=True,
                    )

                    self.exists = True
                    self.location = loc

                    return env
                except Exception as ex:
                    logger.info(
                        f"Failed to create ManagedEnvironment in {loc} due to {ex}"
                    )
            raise ValidationError("Can not find a region with quota to create ManagedEnvironment")

    def get_rid(self):
        rid = self.name
        if not is_valid_resource_id(self.name):
            rid = resource_id(
                subscription=get_subscription_id(self.cmd.cli_ctx),
                resource_group=self.resource_group.name,
                namespace=CONTAINER_APPS_RP,
                type="managedEnvironments",
                name=self.name,
            )
        return rid


class ContainerAppsJob(Resource):  # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            cmd,
            name: str,
            resource_group: "ResourceGroup",
            exists: bool = None,
            image=None,
            env: "ContainerAppEnvironment" = None,
            target_port=None,
            registry_server=None,
            registry_user=None,
            registry_pass=None,
            env_vars=None,
            trigger_type=None,
            replica_timeout=None,
            replica_retry_limit=None,
            replica_completion_count=None,
            parallelism=None,
            cron_expression=None,
    ):

        super().__init__(cmd, name, resource_group, exists)
        self.image = image
        self.env = env
        self.target_port = target_port
        self.registry_server = registry_server
        self.registry_user = registry_user
        self.registry_pass = registry_pass
        self.env_vars = env_vars
        self.trigger_type = trigger_type
        self.replica_timeout = replica_timeout
        self.replica_retry_limit = replica_retry_limit
        self.replica_completion_count = replica_completion_count
        self.parallelism = parallelism
        self.cron_expression = cron_expression
        self.should_create_acr = False
        self.acr: "AzureContainerRegistry" = None

    def _get(self):
        return ContainerAppsJobClient.show(self.cmd, self.resource_group.name, self.name)

    def create(self, no_registry=False):
        # no_registry: don't pass in a registry during create even if the app has one (used for GH actions)
        if get_containerapps_job_if_exists(self.cmd, self.resource_group.name, self.name):
            logger.warning(
                f"Updating Containerapps job {self.name} in resource group {self.resource_group.name}"
            )
        else:
            logger.warning(
                f"Creating Containerapps job {self.name} in resource group {self.resource_group.name}"
            )

        return create_containerappsjob(
            cmd=self.cmd,
            name=self.name,
            resource_group_name=self.resource_group.name,
            image=self.image,
            managed_env=self.env.get_rid(),
            registry_server=None if no_registry else self.registry_server,
            registry_pass=None if no_registry else self.registry_pass,
            registry_user=None if no_registry else self.registry_user,
            env_vars=self.env_vars,
            trigger_type=self.trigger_type,
            replica_timeout=self.replica_timeout,
            replica_retry_limit=self.replica_retry_limit,
            replica_completion_count=self.replica_completion_count,
            parallelism=self.parallelism,
            cron_expression=self.cron_expression
        )


class AzureContainerRegistry(Resource):
    def __init__(self, name: str, resource_group: "ResourceGroup"):  # pylint: disable=super-init-not-called

        self.name = name
        self.resource_group = resource_group


class ContainerApp(Resource):  # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            cmd,
            name: str,
            resource_group: "ResourceGroup",
            exists: bool = None,
            image=None,
            env: "ContainerAppEnvironment" = None,
            target_port=None,
            registry_server=None,
            registry_user=None,
            registry_pass=None,
            env_vars=None,
            workload_profile_name=None,
            ingress=None,
    ):

        super().__init__(cmd, name, resource_group, exists)
        self.image = image
        self.env = env
        self.target_port = target_port
        self.registry_server = registry_server
        self.registry_user = registry_user
        self.registry_pass = registry_pass
        self.env_vars = env_vars
        self.ingress = ingress
        self.workload_profile_name = workload_profile_name

        self.should_create_acr = False
        self.acr: "AzureContainerRegistry" = None

    def _get(self):
        return ContainerAppClient.show(self.cmd, self.resource_group.name, self.name)

    def create(self, no_registry=False):
        # no_registry: don't pass in a registry during create even if the app has one (used for GH actions)
        if get_container_app_if_exists(self.cmd, self.resource_group.name, self.name):
            logger.warning(
                f"Updating Containerapp {self.name} in resource group {self.resource_group.name}"
            )
        else:
            logger.warning(
                f"Creating Containerapp {self.name} in resource group {self.resource_group.name}"
            )

        return containerapp_up_logic(
            cmd=self.cmd,
            name=self.name,
            resource_group_name=self.resource_group.name,
            image=self.image,
            managed_env=self.env.get_rid(),
            target_port=self.target_port,
            registry_server=None if no_registry else self.registry_server,
            registry_pass=None if no_registry else self.registry_pass,
            registry_user=None if no_registry else self.registry_user,
            env_vars=self.env_vars,
            workload_profile_name=self.workload_profile_name,
            ingress=self.ingress,
        )

    def create_acr_if_needed(self):
        if self.should_create_acr:
            logger.warning(
                f"Creating Azure Container Registry {self.acr.name} in resource group "
                f"{self.acr.resource_group.name}"
            )
            self.create_acr()

    def create_acr(self):
        registry_rg = self.resource_group
        url = self.registry_server
        registry_name = url[: url.rindex(ACR_IMAGE_SUFFIX)]
        location = "eastus"
        if self.env.location and self.env.location.lower() != "northcentralusstage":
            location = self.env.location
        registry_def = create_new_acr(
            self.cmd, registry_name, registry_rg.name, location
        )
        self.registry_server = registry_def.login_server

        if not self.acr:
            self.acr = AzureContainerRegistry(registry_name, registry_rg)

        self.registry_user, self.registry_pass, _ = _get_acr_cred(
            self.cmd.cli_ctx, registry_name
        )

    def _docker_push_to_container_registry(self, image_name, forced_acr_login=False):
        from azure.cli.command_modules.acr.custom import acr_login
        from azure.cli.core.profiles import ResourceType

        command = ['docker', 'push', image_name]
        logger.debug(f"Calling '{' '.join(command)}'")
        logger.warning(f"Built image {image_name} locally using buildpacks, attempting to push to registry...")
        try:
            with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
                _, stderr = process.communicate()
                if process.returncode != 0:
                    docker_push_error = stderr.decode('utf-8')
                    if not forced_acr_login and ".azurecr.io/" in image_name and "unauthorized: authentication required" in docker_push_error:
                        # Couldn't push to ACR because the user isn't authenticated. Let's try to login to ACR and retrigger the docker push
                        logger.warning(
                            f"The current user isn't authenticated to the {self.acr.name} ACR instance. Triggering an ACR login and retrying to push the image...")
                        # Logic to login to ACR
                        task_command_kwargs = {"resource_type": ResourceType.MGMT_CONTAINERREGISTRY,
                                               'operation_group': 'webhooks'}
                        old_command_kwargs = {}
                        for key in task_command_kwargs:  # pylint: disable=consider-using-dict-items
                            old_command_kwargs[key] = self.cmd.command_kwargs.get(key)
                            self.cmd.command_kwargs[key] = task_command_kwargs[key]
                        if self.acr and self.acr.name is not None:
                            acr_login(self.cmd, self.acr.name)
                        for k, v in old_command_kwargs.items():
                            self.cmd.command_kwargs[k] = v

                        self._docker_push_to_container_registry(image_name, True)
                    else:
                        raise CLIError(f"Error thrown when running 'docker push': {docker_push_error}")
                logger.debug(f"Successfully pushed image {image_name} to the container registry.")
        except Exception as ex:
            raise CLIError(f"Unable to run 'docker push' command to push image to the container registry: {ex}") from ex

    def build_container_from_source_with_buildpack(self, image_name, source):  # pylint: disable=too-many-statements
        # Ensure that Docker is running
        if not is_docker_running():
            raise ValidationError("Docker is not running. Please start Docker to use buildpacks.")

        # Ensure that the pack CLI is installed
        pack_exec_path = get_pack_exec_path()
        if pack_exec_path is None:
            raise ValidationError("The pack CLI could not be installed.")

        logger.info(
            "Docker is running and pack CLI is installed; attempting to use buildpacks to build container image...")

        registry_name = self.registry_server.lower()
        image_name = f"{registry_name}/{image_name}"
        builder_image_name = "mcr.microsoft.com/oryx/builder:builder-dotnet-7.0"

        # Ensure that the builder is trusted
        command = [pack_exec_path, 'config', 'default-builder', builder_image_name]
        logger.debug(f"Calling '{' '.join(command)}'")
        try:
            with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
                _, stderr = process.communicate()
                if process.returncode != 0:
                    raise CLIError(f"Error thrown when running 'pack config': {stderr.decode('utf-8')}")
                logger.debug(f"Successfully set the default builder to {builder_image_name}.")
        except Exception as ex:
            raise ValidationError(f"Unable to run 'pack config' command to set default builder: {ex}") from ex

        # Run 'pack build' to produce a runnable application image for the Container App
        command = [pack_exec_path, 'build', image_name, '--builder', builder_image_name, '--path', source]
        buildpack_run_tag = get_latest_buildpack_run_tag("aspnet", "7.0")
        if buildpack_run_tag is not None:
            buildpack_run_image = f"mcr.microsoft.com/oryx/builder:{buildpack_run_tag}"
            logger.debug(f"Determined the run image to use as {buildpack_run_image}.")
            command.extend(['--run-image', buildpack_run_image])

        # If the user specifies a target port, pass it to the buildpack
        if self.target_port:
            command.extend(['--env', f"PORT={self.target_port}"])

        logger.debug(f"Calling '{' '.join(command)}'")
        try:
            is_non_supported_platform = False
            with subprocess.Popen(command, stdout=subprocess.PIPE) as process:

                # Stream output of 'pack build' to warning stream
                while process.stdout.readable():
                    line = process.stdout.readline()
                    if not line:
                        break

                    stdout_line = str(line.strip(), 'utf-8')
                    logger.warning(stdout_line)
                    if not is_non_supported_platform and "No buildpack groups passed detection" in stdout_line:
                        is_non_supported_platform = True

                # Update the result of process.returncode
                process.communicate()
                if is_non_supported_platform:
                    raise ValidationError(
                        "Current buildpacks do not support the platform targeted in the provided source code.")

                if process.returncode != 0:
                    raise CLIError(
                        "Non-zero exit code returned from 'pack build'; please check the above output for more details.")

                logger.debug(f"Successfully built image {image_name} using buildpacks.")
        except ValidationError as ex:
            raise ex
        except Exception as ex:
            raise CLIError(f"Unable to run 'pack build' command to produce runnable application image: {ex}") from ex

        # Run 'docker push' to push the image to the container registry
        self._docker_push_to_container_registry(image_name, False)

    def build_container_from_source_with_acr_task(self, image_name, source):
        from azure.cli.command_modules.acr.task import acr_task_create, acr_task_run
        from azure.cli.command_modules.acr._client_factory import cf_acr_tasks, cf_acr_runs
        from azure.cli.core.profiles import ResourceType
        import os

        task_name = "cli_build_containerapp"
        registry_name = (self.registry_server[: self.registry_server.rindex(ACR_IMAGE_SUFFIX)]).lower()
        if not self.target_port:
            self.target_port = DEFAULT_PORT
        task_content = ACR_TASK_TEMPLATE.replace("{{image_name}}", image_name).replace("{{target_port}}",
                                                                                       str(self.target_port))
        task_client = cf_acr_tasks(self.cmd.cli_ctx)
        run_client = cf_acr_runs(self.cmd.cli_ctx)
        task_command_kwargs = {"resource_type": ResourceType.MGMT_CONTAINERREGISTRY, 'operation_group': 'webhooks'}
        old_command_kwargs = {}
        for key in task_command_kwargs:  # pylint: disable=consider-using-dict-items
            old_command_kwargs[key] = self.cmd.command_kwargs.get(key)
            self.cmd.command_kwargs[key] = task_command_kwargs[key]

        with NamedTemporaryFile(mode="w", delete=False) as task_file:
            try:
                task_file.write(task_content)
                task_file.flush()
                acr_task_create(self.cmd, task_client, task_name, registry_name, context_path="/dev/null",
                                file=task_file.name)
                logger.warning("Created ACR task %s in registry %s", task_name, registry_name)
            finally:
                task_file.close()

            from time import sleep
            sleep(10)

            logger.warning("Running ACR build...")
            try:
                acr_task_run(self.cmd, run_client, task_name, registry_name, file=task_file.name, context_path=source)
            except CLIError as e:
                logger.error("Failed to automatically generate a docker container from your source. \n"
                             "See the ACR logs above for more error information. \nPlease check the supported languages for autogenerating docker containers (https://github.com/microsoft/Oryx/blob/main/doc/supportedRuntimeVersions.md), "
                             "or consider using a Dockerfile for your app.")
                raise e
            finally:
                os.unlink(task_file.name)

        for k, v in old_command_kwargs.items():
            self.cmd.command_kwargs[k] = v

    def run_acr_build(self, dockerfile, source, quiet=False, build_from_source=False):
        image_name = self.image if self.image is not None else self.name
        from datetime import datetime

        # Creating a tag for the image using the current time to avoid overwriting customer's existing images
        now = datetime.now()
        tag_now_suffix = str(now).replace(" ", "").replace("-", "").replace(".", "").replace(":", "")
        image_name_with_tag = image_name + ":{}".format(tag_now_suffix)
        self.image = self.registry_server + "/" + image_name_with_tag

        if build_from_source:
            logger.warning(
                "No dockerfile detected. Attempting to build a container directly from the provided source...")

            try:
                # First try to build source using buildpacks
                # Temporary fix: using run time tag as customer image tag
                # Waiting for buildpacks side to fix this issue: https://github.com/buildpacks/pack/issues/1750
                logger.warning("Attempting to build image using buildpacks...")
                buildpack_image_name_with_tag = image_name_with_tag
                run_image_tag = get_latest_buildpack_run_tag("aspnet", "7.0")
                if run_image_tag is not None:
                    buildpack_image_name_with_tag = f"{image_name}:{run_image_tag}-{tag_now_suffix}"
                self.build_container_from_source_with_buildpack(buildpack_image_name_with_tag, source)
                self.image = self.registry_server + "/" + buildpack_image_name_with_tag
                return
            except ValidationError as e:
                logger.warning(f"Unable to use buildpacks to build image from source: {e}\nFalling back to ACR Task...")
            except CLIError as e:
                logger.error("Failed to use buildpacks to build image from source.")
                raise e

            # If we're unable to use the buildpack, build source using an ACR Task
            # Moving tagging img to here
            # Skipping the buildpacks scenario for now due to issues with buildpacks
            # Add version tag for acr image
            logger.warning("Attempting to build image using ACR Task...")
            self.build_container_from_source_with_acr_task(image_name_with_tag, source)
        else:
            # Moving tagging img to here
            # Skipping the buildpacks scenario for now due to issues with buildpacks
            # Add version tag for acr image
            queue_acr_build(
                self.cmd,
                self.acr.resource_group.name,
                self.acr.name,
                image_name_with_tag,
                source,
                dockerfile,
                quiet,
            )


def _create_service_principal(cmd, resource_group_name, env_resource_group_name):
    logger.warning(
        "No valid service principal provided. Creating a new service principal..."
    )
    scopes = [
        f"/subscriptions/{get_subscription_id(cmd.cli_ctx)}/resourceGroups/{resource_group_name}"
    ]
    if env_resource_group_name is not None and env_resource_group_name != resource_group_name:
        scopes.append(
            f"/subscriptions/{get_subscription_id(cmd.cli_ctx)}/resourceGroups/{env_resource_group_name}"
        )
    sp = create_service_principal_for_github_action(cmd, scopes=scopes, role="contributor")

    logger.warning(f"Created service principal with ID {sp['appId']}")

    return sp["appId"], sp["password"], sp["tenant"]


def _get_or_create_sp(  # pylint: disable=inconsistent-return-statements
        cmd,
        resource_group_name,
        env_resource_group_name,
        name,
        service_principal_client_id,
        service_principal_client_secret,
        service_principal_tenant_id,
):
    if service_principal_client_id and service_principal_client_secret and service_principal_tenant_id:
        return (
            service_principal_client_id,
            service_principal_client_secret,
            service_principal_tenant_id,
        )
    try:
        GitHubActionClient.show(
            cmd=cmd, resource_group_name=resource_group_name, name=name
        )
        return (
            service_principal_client_id,
            service_principal_client_secret,
            service_principal_tenant_id,
        )
    except:  # pylint: disable=bare-except
        service_principal = None

        # TODO if possible, search for SPs with the right credentials
        # I haven't found a way to get SP creds + secrets yet from the API

        if not service_principal:
            return _create_service_principal(
                cmd, resource_group_name, env_resource_group_name
            )
        # return client_id, secret, tenant_id


def _get_dockerfile_content_from_repo(
        repo_url, branch, token, context_path, dockerfile
):
    from github import Github

    g = Github(token)
    context_path = context_path or "."
    repo = repo_url_to_name(repo_url)
    try:
        r = g.get_repo(repo)
        if not branch:
            branch = r.default_branch
    except Exception as e:
        raise ValidationError(f"Could not find repo {repo_url}") from e
    try:
        files = r.get_contents(context_path, ref=branch)
    except Exception as e:
        raise ValidationError(f"Could not find branch {branch}") from e
    for f in files:
        if f.path == dockerfile or f.path.endswith(f"/{dockerfile}"):
            resp = requests.get(f.download_url)
            if resp.ok and resp.content:
                return resp.content.decode("utf-8").split("\n")
    raise ValidationError("Could not find Dockerfile in Github repo/branch. Please ensure it is named 'Dockerfile'. "
                          "Set the path with --context-path if not in the root directory.")


def _get_ingress_and_target_port(ingress, target_port, dockerfile_content: "list[str]"):
    if not target_port and not ingress and dockerfile_content is not None:  # pylint: disable=too-many-nested-blocks
        for line in dockerfile_content:
            if line:
                line = (
                    line.upper()
                    .strip()
                    .replace("/TCP", "")
                    .replace("/UDP", "")
                    .replace("\n", "")
                )
                if line and line[0] != "#":
                    if "EXPOSE" in line:
                        parts = line.split(" ")
                        for i, p in enumerate(parts[:-1]):
                            if "EXPOSE" in p and is_int(parts[i + 1]):
                                target_port = parts[i + 1]
                                ingress = "external"
                                logger.warning(
                                    "Adding external ingress port {} based on dockerfile expose.".format(
                                        target_port
                                    )
                                )
    ingress = "external" if target_port and not ingress else ingress
    return ingress, target_port


def _validate_up_args(cmd, source, image, repo, registry_server):
    disallowed_params = ["--only-show-errors", "--output", "-o"]
    command_args = cmd.cli_ctx.data.get("safe_params", [])
    for a in disallowed_params:
        if a in command_args:
            raise ValidationError(f"Argument {a} is not allowed for 'az containerapp up'")

    if not source and not image and not repo:
        raise RequiredArgumentMissingError(
            "You must specify either --source, --repo, or --image"
        )
    if source and repo:
        raise MutuallyExclusiveArgumentError(
            "Cannot use --source and --repo togther. "
            "Can either deploy from a local directory or a Github repo"
        )
    if repo and registry_server and "azurecr.io" in registry_server:
        parsed = urlparse(registry_server)
        registry_name = (parsed.netloc if parsed.scheme else parsed.path).split(".")[0]
        if registry_name and len(registry_name) > MAXIMUM_SECRET_LENGTH:
            raise ValidationError(f"--registry-server ACR name must be less than {MAXIMUM_SECRET_LENGTH} "
                                  "characters when using --repo")


def _reformat_image(source, repo, image):
    if source and (image or repo):
        image = image.split("/")[-1]  # if link is given
        image = image.replace(":", "")
    return image


def _has_dockerfile(source, dockerfile):
    try:
        content = _get_dockerfile_content_local(source, dockerfile)
        return bool(content)
    except InvalidArgumentValueError:
        return False


def _get_dockerfile_content_local(source, dockerfile):
    lines = []
    if source:
        dockerfile_location = f"{source}/{dockerfile}"
        try:
            with open(dockerfile_location, "r") as fh:  # pylint: disable=unspecified-encoding
                lines = list(fh)
        except Exception as e:
            raise InvalidArgumentValueError(
                "Cannot open specified Dockerfile. Check dockerfile name, path, and permissions."
            ) from e
    return lines


def _get_dockerfile_content(repo, branch, token, source, context_path, dockerfile):
    if source:
        return _get_dockerfile_content_local(source, dockerfile)
    elif repo:
        return _get_dockerfile_content_from_repo(
            repo, branch, token, context_path, dockerfile
        )
    return []


def _get_app_env_and_group(
        cmd, name, resource_group: "ResourceGroup", env: "ContainerAppEnvironment", location
):
    if not resource_group.name and not resource_group.exists:
        matched_apps = [c for c in list_containerapp(cmd) if c["name"].lower() == name.lower()]
        if env.name:
            matched_apps = [c for c in matched_apps if
                            parse_resource_id(c["properties"]["environmentId"])["name"].lower() == env.name.lower()]
        if location:
            matched_apps = [c for c in matched_apps if format_location(c["location"]) == format_location(location)]
        if len(matched_apps) == 1:
            resource_group.name = parse_resource_id(matched_apps[0]["id"])[
                "resource_group"
            ]
            env.set_name(matched_apps[0]["properties"]["environmentId"])
        elif len(matched_apps) > 1:
            raise ValidationError(
                f"There are multiple containerapps with name {name} on the subscription. "
                "Please specify which resource group your Containerapp is in."
            )


def _get_env_and_group_from_log_analytics(
        cmd,
        resource_group_name,
        env: "ContainerAppEnvironment",
        resource_group: "ResourceGroup",
        logs_customer_id,
        location,
):
    # resource_group_name is the value the user passed in (if present)
    if not env.name:
        if (resource_group_name == resource_group.name and resource_group.exists) or (
                not resource_group_name
        ):
            env_list = list_managed_environments(
                cmd=cmd, resource_group_name=resource_group_name
            )
            if logs_customer_id:
                env_list = [
                    e
                    for e in env_list
                    if safe_get(e, "properties", "appLogsConfiguration", "logAnalyticsConfiguration", "customerId") == logs_customer_id
                ]
            if location:
                env_list = [e for e in env_list if format_location(e["location"]) == format_location(location)]
            if env_list:
                # TODO check how many CA in env
                env_details = parse_resource_id(env_list[0]["id"])
                env.set_name(env_details["name"])
                resource_group.name = env_details["resource_group"]


def _get_acr_from_image(cmd, app):
    if app.image is not None and "azurecr.io" in app.image:
        app.registry_server = app.image.split("/")[
            0
        ]  # TODO what if this conflicts with registry_server param?
        parsed = urlparse(app.image)
        registry_name = (parsed.netloc if parsed.scheme else parsed.path).split(".")[0]
        if app.registry_user is None or app.registry_pass is None:
            logger.info(
                "No credential was provided to access Azure Container Registry. Trying to look up..."
            )
            try:
                app.registry_user, app.registry_pass, registry_rg = _get_acr_cred(
                    cmd.cli_ctx, registry_name
                )
                app.acr = AzureContainerRegistry(
                    registry_name, ResourceGroup(cmd, registry_rg, None, None)
                )
            except Exception as ex:
                raise RequiredArgumentMissingError(
                    "Failed to retrieve credentials for container registry. Please provide the registry username and password"
                ) from ex
        else:
            acr_rg = _get_acr_rg(app)
            app.acr = AzureContainerRegistry(
                name=registry_name,
                resource_group=ResourceGroup(app.cmd, acr_rg, None, None),
            )


def _get_registry_from_app(app, source):
    containerapp_def = app.get()
    existing_registries = safe_get(containerapp_def, "properties", "configuration", "registries", default=[])
    if source:
        existing_registries = [r for r in existing_registries if ACR_IMAGE_SUFFIX in r["server"]]
    if containerapp_def:
        if len(existing_registries) == 1:
            app.registry_server = existing_registries[0]["server"]
        elif len(existing_registries) > 1:  # default to registry in image if possible, otherwise don't infer
            containers = safe_get(containerapp_def, "properties", "template", "containers", default=[])
            image_server = next(c["image"] for c in containers if c["name"].lower() == app.name.lower()).split('/')[0]
            if image_server in [r["server"] for r in existing_registries]:
                app.registry_server = image_server


def _get_acr_rg(app):
    registry_name = app.registry_server[: app.registry_server.rindex(ACR_IMAGE_SUFFIX)]
    client = get_mgmt_service_client(
        app.cmd.cli_ctx, ContainerRegistryManagementClient
    ).registries
    return parse_resource_id(acr_show(app.cmd, client, registry_name).id)[
        "resource_group"
    ]


def _get_default_registry_name(app):
    import hashlib

    h = hashlib.sha256()
    h.update(f"{get_subscription_id(app.cmd.cli_ctx)}/{app.env.resource_group.name}/{app.env.name}".encode("utf-8"))

    registry_name = f"{h.hexdigest()}"[:10]  # cap at 15 characters total
    return f"ca{registry_name}acr"  # ACR names must start + end in a letter


def _set_acr_creds(cmd, app: "ContainerApp", registry_name):
    logger.info("No credential was provided to access Azure Container Registry. Trying to look up...")
    try:
        app.registry_user, app.registry_pass, registry_rg = _get_acr_cred(
            cmd.cli_ctx, registry_name
        )
        return registry_rg
    except Exception as ex:
        raise RequiredArgumentMissingError(
            "Failed to retrieve credentials for container registry. Please provide the registry username and password"
        ) from ex


def _get_registry_details(cmd, app: "ContainerApp", source):
    registry_rg = None
    registry_name = None
    if app.registry_server:
        if "azurecr.io" not in app.registry_server and source:
            raise ValidationError(
                "Cannot supply non-Azure registry when using --source."
            )
        parsed = urlparse(app.registry_server)
        registry_name = (parsed.netloc if parsed.scheme else parsed.path).split(".")[0]
        if app.registry_user is None or app.registry_pass is None:
            registry_rg = _set_acr_creds(cmd, app, registry_name)
        else:
            registry_rg = _get_acr_rg(app)
    else:
        registry_name, registry_rg = find_existing_acr(cmd, app)
        if registry_name and registry_rg:
            _set_acr_creds(cmd, app, registry_name)
            app.registry_server = registry_name + ACR_IMAGE_SUFFIX
        else:
            registry_rg = app.resource_group.name
            registry_name = _get_default_registry_name(app)
            app.registry_server = registry_name + ACR_IMAGE_SUFFIX
            app.should_create_acr = True

    app.acr = AzureContainerRegistry(
        registry_name, ResourceGroup(cmd, registry_rg, None, None)
    )


# attempt to populate defaults for managed env, RG, ACR, etc
def _set_up_defaults(
        cmd,
        name,
        resource_group_name,
        logs_customer_id,
        location,
        resource_group: "ResourceGroup",
        env: "ContainerAppEnvironment",
        app: "ContainerApp",
):
    # If no RG passed in and a singular app exists with the same name, get its env and rg
    _get_app_env_and_group(cmd, name, resource_group, env, location)

    # If no env passed in (and not creating a new RG), then try getting an env by location / log analytics ID
    _get_env_and_group_from_log_analytics(
        cmd, resource_group_name, env, resource_group, logs_customer_id, location
    )

    # try to set RG name by env name
    if env.name and not resource_group.name:
        if not location:
            env_list = [e for e in list_managed_environments(cmd=cmd) if e["name"] == env.name]
        else:
            env_list = [e for e in list_managed_environments(cmd=cmd) if
                        e["name"] == env.name and format_location(e["location"]) == format_location(location)]
        if len(env_list) == 1:
            resource_group.name = parse_resource_id(env_list[0]["id"])["resource_group"]
        if len(env_list) > 1:
            raise ValidationError(
                f"There are multiple environments with name {env.name} on the subscription. "
                "Please specify which resource group your Containerapp environment is in."
            )  # get ACR details from --image, if possible
    _get_acr_from_image(cmd, app)


def _create_github_action(
        app: "ContainerApp",
        env: "ContainerAppEnvironment",
        service_principal_client_id,
        service_principal_client_secret,
        service_principal_tenant_id,
        branch,
        token,
        repo,
        context_path,
):
    sp = _get_or_create_sp(
        app.cmd,
        app.resource_group.name,
        env.resource_group.name,
        app.name,
        service_principal_client_id,
        service_principal_client_secret,
        service_principal_tenant_id,
    )
    (
        service_principal_client_id,
        service_principal_client_secret,
        service_principal_tenant_id,
    ) = sp

    create_or_update_github_action(
        cmd=app.cmd,
        name=app.name,
        resource_group_name=app.resource_group.name,
        repo_url=repo,
        registry_url=app.registry_server,
        registry_username=app.registry_user,
        registry_password=app.registry_pass,
        branch=branch,
        token=token,
        login_with_github=False,
        service_principal_client_id=service_principal_client_id,
        service_principal_client_secret=service_principal_client_secret,
        service_principal_tenant_id=service_principal_tenant_id,
        image=app.image,
        context_path=context_path,
        trigger_existing_workflow=True,
    )


def up_output(app: 'ContainerApp', no_dockerfile):
    url = safe_get(
        ContainerAppClient.show(app.cmd, app.resource_group.name, app.name),
        "properties",
        "configuration",
        "ingress",
        "fqdn",
    )
    if url and not url.startswith("http"):
        url = f"http://{url}"

    logger.warning(
        f"\nYour container app {app.name} has been created and deployed! Congrats! \n"
    )
    if no_dockerfile and app.ingress:
        logger.warning(f"Your app is running image {app.image} and listening on port {app.target_port}")

    url and logger.warning(f"Browse to your container app at: {url} \n")
    logger.warning(
        f"Stream logs for your container with: az containerapp logs show -n {app.name} -g {app.resource_group.name} \n"
    )
    logger.warning(
        f"See full output using: az containerapp show -n {app.name} -g {app.resource_group.name} \n"
    )


def find_existing_acr(cmd, app: "ContainerApp"):
    from azure.cli.command_modules.acr._client_factory import cf_acr_registries
    client = cf_acr_registries(cmd.cli_ctx)

    acr = None
    try:
        acr = acr_show(cmd, client=client, registry_name=_get_default_registry_name(app))
    except Exception:
        pass

    if acr:
        app.should_create_acr = False
        return acr.name, parse_resource_id(acr.id)["resource_group"]
    return None, None


def check_env_name_on_rg(cmd, managed_env, resource_group_name, location):
    if location:
        _ensure_location_allowed(cmd, location, CONTAINER_APPS_RP, "managedEnvironments")
    if managed_env and resource_group_name and location:
        env_def = None
        try:
            env_def = ManagedEnvironmentClient.show(cmd, resource_group_name, parse_resource_id(managed_env)["name"])
        except:  # pylint: disable=bare-except
            pass
        if env_def:
            if format_location(location) != format_location(env_def["location"]):
                raise ValidationError(
                    "Environment {} already exists in resource group {} on location {}, cannot change location of existing environment to {}.".format(
                        parse_resource_id(managed_env)["name"], resource_group_name, env_def["location"], location))


def get_token(cmd, repo, token):
    if not repo:
        return None
    if token:
        return token
    token = load_github_token_from_cache(cmd, repo)
    if not token:
        token = get_github_access_token(cmd, ["admin:repo_hook", "repo", "workflow"], token)
    return token



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: skip-file
# flake8: noqa
from msrest.serialization import Model


class AdministratorDetails(Model):
    """Details of the organization administrator of the certificate issuer.

    :param first_name: First name.
    :type first_name: str
    :param last_name: Last name.
    :type last_name: str
    :param email_address: Email addresss.
    :type email_address: str
    :param phone: Phone number.
    :type phone: str
    """

    _attribute_map = {
        'first_name': {'key': 'first_name', 'type': 'str'},
        'last_name': {'key': 'last_name', 'type': 'str'},
        'email_address': {'key': 'email', 'type': 'str'},
        'phone': {'key': 'phone', 'type': 'str'},
    }

    def __init__(self, **kwargs):
        super(AdministratorDetails, self).__init__(**kwargs)
        self.first_name = kwargs.get('first_name', None)
        self.last_name = kwargs.get('last_name', None)
        self.email_address = kwargs.get('email_address', None)
        self.phone = kwargs.get('phone', None)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long, unused-argument

import re
from azure.cli.core.azclierror import (ValidationError, ResourceNotFoundError, InvalidArgumentValueError,
                                       MutuallyExclusiveArgumentError)
from msrestazure.tools import is_valid_resource_id
from knack.log import get_logger

from ._clients import ContainerAppClient
from ._ssh_utils import ping_container_app
from ._utils import safe_get, is_registry_msi_system
from ._constants import ACR_IMAGE_SUFFIX, LOG_TYPE_SYSTEM


logger = get_logger(__name__)


# called directly from custom method bc otherwise it disrupts the --environment auto RID functionality
def validate_create(registry_identity, registry_pass, registry_user, registry_server, no_wait):
    if registry_identity and (registry_pass or registry_user):
        raise MutuallyExclusiveArgumentError("Cannot provide both registry identity and username/password")
    if is_registry_msi_system(registry_identity) and no_wait:
        raise MutuallyExclusiveArgumentError("--no-wait is not supported with system registry identity")
    if registry_identity and not is_valid_resource_id(registry_identity) and not is_registry_msi_system(registry_identity):
        raise InvalidArgumentValueError("--registry-identity must be an identity resource ID or 'system'")
    if registry_identity and ACR_IMAGE_SUFFIX not in (registry_server or ""):
        raise InvalidArgumentValueError("--registry-identity: expected an ACR registry (*.azurecr.io) for --registry-server")


def _is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def validate_revision_suffix(value):
    if value is not None:
        # what does the following regex check?
        # 1. ^[a-z0-9] - starts with a letter or number
        # 2. (?!.*-{2}) - does not contain '--'
        # 3. ([-a-z0-9]*[a-z0-9])? - ends with a letter or number and can contain '-' in between
        matched = re.match(r"^[a-z0-9](?!.*-{2})([-a-z0-9]*[a-z0-9])?$", value)
        if not matched:
            raise ValidationError(f"Invalid Container App revision suffix '{value}'. A revision suffix must consist of lower case alphanumeric characters or '-', start with a letter or number, end with an alphanumeric character and cannot have '--'.")


def validate_memory(namespace):
    if namespace.memory is not None:
        valid = False

        if not namespace.memory.endswith("Gi"):
            namespace.memory = namespace.memory.rstrip()
            namespace.memory += "Gi"

        valid = _is_number(namespace.memory[:-2])

        if not valid:
            raise ValidationError("Usage error: --memory must be a number ending with \"Gi\"")


def validate_cpu(namespace):
    if namespace.cpu:
        cpu = namespace.cpu
        try:
            float(cpu)
        except ValueError as e:
            raise ValidationError("Usage error: --cpu must be a number eg. \"0.5\"") from e


def validate_managed_env_name_or_id(cmd, namespace):
    from azure.cli.core.commands.client_factory import get_subscription_id
    from msrestazure.tools import resource_id

    if namespace.managed_env:
        if not is_valid_resource_id(namespace.managed_env):
            namespace.managed_env = resource_id(
                subscription=get_subscription_id(cmd.cli_ctx),
                resource_group=namespace.resource_group_name,
                namespace='Microsoft.App',
                type='managedEnvironments',
                name=namespace.managed_env
            )


def validate_storage_name_or_id(cmd, namespace):
    from azure.cli.core.commands.client_factory import get_subscription_id
    from msrestazure.tools import resource_id

    if namespace.storage_account:
        if not is_valid_resource_id(namespace.storage_account):
            namespace.storage_account = resource_id(
                subscription=get_subscription_id(cmd.cli_ctx),
                resource_group=namespace.resource_group_name,
                namespace='Microsoft.Storage',
                type='storageAccounts',
                name=namespace.storage_account
            )


def validate_registry_server(namespace):
    if "create" in namespace.command.lower():
        if namespace.registry_server:
            if not namespace.registry_user or not namespace.registry_pass:
                if ACR_IMAGE_SUFFIX not in namespace.registry_server:
                    raise ValidationError("Usage error: --registry-server, --registry-password and --registry-username are required together if not using Azure Container Registry")


def validate_registry_user(namespace):
    if "create" in namespace.command.lower():
        if namespace.registry_user:
            if not namespace.registry_server or (not namespace.registry_pass and ACR_IMAGE_SUFFIX not in namespace.registry_server):
                raise ValidationError("Usage error: --registry-server, --registry-password and --registry-username are required together if not using Azure Container Registry")


def validate_registry_pass(namespace):
    if "create" in namespace.command.lower():
        if namespace.registry_pass:
            if not namespace.registry_server or (not namespace.registry_user and ACR_IMAGE_SUFFIX not in namespace.registry_server):
                raise ValidationError("Usage error: --registry-server, --registry-password and --registry-username are required together if not using Azure Container Registry")


def validate_target_port(namespace):
    if "create" in namespace.command.lower():
        if namespace.target_port:
            if not namespace.ingress:
                raise ValidationError("Usage error: must specify --ingress with --target-port")


def validate_ingress(namespace):
    if "create" in namespace.command.lower():
        if namespace.ingress:
            if not namespace.target_port:
                raise ValidationError("Usage error: must specify --target-port with --ingress")


def validate_allow_insecure(namespace):
    if "create" in namespace.command.lower():
        if namespace.allow_insecure:
            if not namespace.ingress or not namespace.target_port:
                raise ValidationError("Usage error: must specify --ingress and --target-port with --allow-insecure")
            if namespace.transport == "tcp":
                raise ValidationError("Usage error: --allow-insecure is not supported for TCP ingress")


def _set_ssh_defaults(cmd, namespace):
    app = ContainerAppClient.show(cmd, namespace.resource_group_name, namespace.name)
    if not app:
        raise ResourceNotFoundError("Could not find a container app")
    replicas = []
    if not namespace.revision:
        namespace.revision = app.get("properties", {}).get("latestRevisionName")
        if not namespace.revision:
            raise ResourceNotFoundError("Could not find a revision")
    if not namespace.replica:
        # VVV this may not be necessary according to Anthony Chu
        try:
            ping_container_app(app)  # needed to get an alive replica
        except Exception as e:  # pylint: disable=broad-except
            logger.warning("Failed to ping container app with error '%s' \nPlease ensure there is an alive replica. ", str(e))
        replicas = ContainerAppClient.list_replicas(cmd=cmd,
                                                    resource_group_name=namespace.resource_group_name,
                                                    container_app_name=namespace.name,
                                                    revision_name=namespace.revision)
        if not replicas:
            raise ResourceNotFoundError("Could not find a replica for this app")
        namespace.replica = replicas[0]["name"]
    if not namespace.container:
        revision = ContainerAppClient.show_revision(cmd, resource_group_name=namespace.resource_group_name,
                                                    container_app_name=namespace.name,
                                                    name=namespace.revision)
        revision_containers = safe_get(revision, "properties", "template", "containers")
        if revision_containers:
            namespace.container = revision_containers[0]["name"]


def _validate_revision_exists(cmd, namespace):
    revision = ContainerAppClient.show_revision(cmd, resource_group_name=namespace.resource_group_name,
                                                container_app_name=namespace.name, name=namespace.revision)
    if not revision:
        raise ResourceNotFoundError("Could not find revision")


def _validate_replica_exists(cmd, namespace):
    replica = ContainerAppClient.get_replica(cmd=cmd,
                                             resource_group_name=namespace.resource_group_name,
                                             container_app_name=namespace.name,
                                             revision_name=namespace.revision,
                                             replica_name=namespace.replica)
    if not replica:
        raise ResourceNotFoundError("Could not find replica")


def _validate_container_exists(cmd, namespace):
    replica_containers = ContainerAppClient.get_replica(cmd=cmd,
                                                        resource_group_name=namespace.resource_group_name,
                                                        container_app_name=namespace.name,
                                                        revision_name=namespace.revision,
                                                        replica_name=namespace.replica)["properties"]["containers"]
    matches = [r for r in replica_containers if r["name"].lower() == namespace.container.lower()]
    if not matches:
        raise ResourceNotFoundError("Could not find container")


# also used to validate logstream
def validate_ssh(cmd, namespace):
    if not hasattr(namespace, "kind") or (namespace.kind and namespace.kind.lower() != LOG_TYPE_SYSTEM):
        _set_ssh_defaults(cmd, namespace)
        _validate_revision_exists(cmd, namespace)
        _validate_replica_exists(cmd, namespace)
        _validate_container_exists(cmd, namespace)


def validate_cors_max_age(cmd, namespace):
    if namespace.max_age:
        try:
            if namespace.max_age == "":
                return

            max_age = int(namespace.max_age)
            if max_age < 0:
                raise InvalidArgumentValueError("max-age must be a positive integer.")
        except ValueError:
            raise InvalidArgumentValueError("max-age must be an integer.")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long, protected-access
import json
from knack.log import get_logger
from knack.util import CLIError
from azure.cli.core.aaz import has_value, AAZListArg, AAZStrArg, AAZIntArg, AAZBoolArg
from azure.cli.command_modules.monitor.actions import AAZCustomListArg
from azure.cli.command_modules.monitor._autoscale_util import get_autoscale_default_profile
from ..aaz.latest.monitor.autoscale import Create as _AutoScaleCreate, Update as _AutoScaleUpdate, \
    Show as _AutoScaleShow, List as _AutoScaleList
from azure.cli.command_modules.network.custom import _convert_to_snake_case
from azure.cli.core.azclierror import InvalidArgumentValueError

logger = get_logger(__name__)


DEFAULT_PROFILE_NAME = 'default'


class AutoScaleCreate(_AutoScaleCreate):

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        if has_value(self.ctx.vars.instance.properties.name):
            self.ctx.vars.instance.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result


class AutoScaleShow(_AutoScaleShow):

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        # When the name field conflicts, the name in inner layer is ignored and the outer layer is applied
        if has_value(self.ctx.vars.instance.properties.name):
            self.ctx.vars.instance.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result


def update_add_actions(args):
    add_actions = []
    for add_action_item in args.add_actions:
        add_action_item_arr = add_action_item.to_serialized_data()
        _type = add_action_item_arr[0].lower()
        if _type == "email":
            add_actions.append({
                "key": "email",
                "value": {
                    "customEmails": add_action_item_arr[1:]
                }
            })
        elif _type == "webhook":
            uri = add_action_item_arr[1]
            try:
                properties = dict(x.split('=', 1) for x in add_action_item_arr[2:])
            except ValueError:
                raise InvalidArgumentValueError('webhook URI [KEY=VALUE ...]')
            add_actions.append({
                "key": "webhook",
                "value": {
                    "serviceUri": uri,
                    "properties": properties
                }
            })
        else:
            raise InvalidArgumentValueError('TYPE KEY [ARGS]')
    return add_actions


def update_remove_actions(args):
    remove_actions = []
    for remove_action_item in args.remove_actions:
        values = remove_action_item.to_serialized_data()
        _type = values[0].lower()
        if _type not in ['email', 'webhook']:
            raise InvalidArgumentValueError('TYPE KEY [KEY ...]')
        remove_actions.append(values[1:])
    return remove_actions


class AutoScaleUpdate(_AutoScaleUpdate):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.notifications._registered = False
        args_schema.profiles._registered = False
        args_schema.target_resource_location._registered = False
        args_schema.target_resource_uri._registered = False
        args_schema.count = AAZIntArg(
            options=["--count"],
            help='The numer of instances to use. If used with --min/max-count, the default number of instances to use.',
            arg_group="Instance Limit",
        )
        args_schema.min_count = AAZIntArg(
            options=["--min-count"],
            help='The minimum number of instances.',
            arg_group="Instance Limit",
        )
        args_schema.max_count = AAZIntArg(
            options=["--max-count"],
            help='The maximum number of instances.',
            arg_group="Instance Limit",
        )
        args_schema.add_actions = AAZCustomListArg(
            options=["--add-actions"],
            singular_options=['--add-action', '-a'],
            help="Add an action to fire when a scaling event occurs." + '''
        Usage:   --add-action TYPE KEY [ARG ...]
        Email:   --add-action email bob@contoso.com ann@contoso.com
        Webhook: --add-action webhook https://www.contoso.com/alert apiKey=value
        Webhook: --add-action webhook https://www.contoso.com/alert?apiKey=value
        Multiple actions can be specified by using more than one `--add-action` argument.
        ''',
            arg_group="Notification",
        )
        args_schema.add_actions.Element = AAZCustomListArg()
        args_schema.add_actions.Element.Element = AAZStrArg()

        args_schema.remove_actions = AAZCustomListArg(
            options=["--remove-actions"],
            singular_options=['--remove-action', '-r'],
            help="Remove one or more actions." + '''
        Usage:   --remove-action TYPE KEY [KEY ...]
        Email:   --remove-action email bob@contoso.com ann@contoso.com
        Webhook: --remove-action webhook https://contoso.com/alert https://alerts.contoso.com.
        ''',
            arg_group="Notification",
        )
        args_schema.remove_actions.Element = AAZCustomListArg()
        args_schema.remove_actions.Element.Element = AAZStrArg()

        args_schema.email_administrator = AAZBoolArg(
            options=["--email-administrator"],
            help='Send email to subscription administrator on scaling.',
            arg_group="Notification",
        )
        args_schema.email_coadministrators = AAZBoolArg(
            options=["--email-coadministrators"],
            help='Send email to subscription co-administrators on scaling.',
            arg_group="Notification",
        )

        return args_schema

    def pre_instance_update(self, instance):
        args = self.ctx.args
        add_actions = update_add_actions(args)
        remove_actions = update_remove_actions(args)
        if has_value(args.count) or has_value(args.min_count) or has_value(args.max_count):
            default_profile = get_autoscale_default_profile(instance)
            curr_count = default_profile["capacity"]["default"]
            curr_min = default_profile["capacity"]["minimum"]
            curr_max = default_profile["capacity"]["maximum"]
            is_fixed_count = curr_count == curr_min and curr_count == curr_max

            # check for special case where count is used to indicate fixed value and only
            # count is updated
            if has_value(args.count) and is_fixed_count and not has_value(args.min_count) and not has_value(args.max_count):
                args.min_count = args.count.to_serialized_data()
                args.max_count = args.count.to_serialized_data()

            count = curr_count if not has_value(args.count) else args.count.to_serialized_data()
            min_count = curr_min if not has_value(args.min_count) else args.min_count.to_serialized_data()
            max_count = curr_max if not has_value(args.max_count) else args.max_count.to_serialized_data()

            # There may be multiple "default" profiles. All need to updated.
            for profile in instance.properties.profiles:
                if has_value(profile.fixed_date):
                    continue
                if has_value(profile.recurrence):
                    try:
                        # portal denotes the "default" pairs by using a JSON string for their name
                        # so if it can be decoded, we know this is a default profile
                        json.loads(profile.name.to_serialized_data())
                    except ValueError:
                        continue
                profile.capacity.default = str(count)
                profile.capacity.minimum = str(min_count)
                profile.capacity.maximum = str(max_count)
        updated_notification = None
        if instance.properties.notifications:
            retained_notification = []
            for x in instance.properties.notifications:
                note = x.to_serialized_data()
                if note['operation'].lower() == 'scale':
                    updated_notification = note
                else:
                    retained_notification.append(note)
            instance.properties.notifications = retained_notification
        else:
            instance.properties.notifications = []

        if updated_notification is None:
            updated_notification = {
                "operation": "scale",
                "email": {
                    "customEmails": []
                },
                "webhooks": []
            }

        # process removals
        if len(remove_actions) > 0:
            removed_emails, removed_webhooks = _parse_action_removals(remove_actions)
            updated_notification['email']['customEmails'] = \
                [x for x in updated_notification['email']['customEmails'] if x not in removed_emails]
            updated_notification['webhooks'] = \
                [x for x in updated_notification['webhooks'] if x['serviceUri'] not in removed_webhooks]

        # process additions
        for action in add_actions:
            if action["key"] == "email":
                for email in action["value"]["customEmails"]:
                    updated_notification['email']['customEmails'].append(email)
            elif action["key"] == "webhook":
                updated_notification['webhooks'].append(action["value"])
        if has_value(args.email_administrator):
            updated_notification['email']['sendToSubscriptionAdministrator'] = args.email_administrator.to_serialized_data()
        if has_value(args.email_coadministrators):
            updated_notification['email']['sendToSubscriptionCoAdministrators'] = args.email_coadministrators.to_serialized_data()

        instance.properties.notifications.append(updated_notification)

        if has_value(args.scale_look_ahead_time) and not has_value(args.scale_mode) \
                and not has_value(instance.properties.predictive_autoscale_policy):
            raise InvalidArgumentValueError('scale-mode is required for setting scale-look-ahead-time.')

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        # When the name field conflicts, the name in inner layer is ignored and the outer layer is applied
        if has_value(self.ctx.vars.instance.properties.name):
            self.ctx.vars.instance.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result


class AutoScaleList(_AutoScaleList):

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        # When the name field conflicts, the name in inner layer is ignored and the outer layer is applied
        for value in self.ctx.vars.instance.value:
            if has_value(value.properties):
                value.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance.value, client_flatten=True)
        next_link = self.deserialize_output(self.ctx.vars.instance.next_link)
        return result, next_link


# pylint: disable=too-many-locals
def autoscale_create(cmd, resource, count, autoscale_name=None, resource_group_name=None,
                     min_count=None, max_count=None, location=None, tags=None, disabled=None,
                     actions=None, email_administrator=None, email_coadministrators=None,
                     scale_mode=None, scale_look_ahead_time=None):

    if not autoscale_name:
        from msrestazure.tools import parse_resource_id
        autoscale_name = parse_resource_id(resource)['name']
    min_count = min_count or count
    max_count = max_count or count
    args = {}
    default_profile = {
        "name": DEFAULT_PROFILE_NAME,
        "capacity": {
            "default": str(count),
            "minimum": str(min_count),
            "maximum": str(max_count)
        },
        "rules": []
    }

    notification = {
        "operation": "scale",
        "email": {
            "custom_emails": [],
            "send_to_subscription_administrator": email_administrator,
            "send_to_subscription_co_administrators": email_coadministrators,
        },
        "webhooks": []
    }

    for action in actions or []:
        key = action["key"]
        value = action["value"]
        if key == "email":
            for email in value["custom_emails"]:
                notification["email"]["custom_emails"].append(email)
        elif key == "webhook":
            notification["webhooks"].append(value)

    if scale_mode is not None and scale_look_ahead_time is not None:
        args["scale_mode"] = scale_mode
        args["scale_look_ahead_time"] = scale_look_ahead_time
    elif scale_mode is not None:
        args["scale_mode"] = scale_mode
    elif scale_look_ahead_time is not None:
        raise InvalidArgumentValueError('scale-mode is required for setting predictive autoscale policy.')
    args["location"] = location
    args["profiles"] = [default_profile]
    args["tags"] = tags
    args["notifications"] = [notification]
    args["enabled"] = not disabled
    args["autoscale_name"] = autoscale_name
    args["target_resource_uri"] = resource
    args["resource_group"] = resource_group_name

    if not (min_count == count and max_count == count):
        logger.warning('Follow up with `az monitor autoscale rule create` to add scaling rules.')
    return AutoScaleCreate(cli_ctx=cmd.cli_ctx)(command_args=args)


def _parse_action_removals(actions):
    """ Separates the combined list of keys to remove into webhooks and emails. """
    flattened = list({x for sublist in actions for x in sublist})
    emails = []
    webhooks = []
    for item in flattened:
        if item.startswith('http://') or item.startswith('https://'):
            webhooks.append(item)
        else:
            emails.append(item)
    return emails, webhooks


def _apply_copy_rules(autoscale_settings, new_profile, copy_rules):
    if copy_rules:
        copy_profile = next(x for x in autoscale_settings["profiles"] if x["name"] == copy_rules)
        if copy_profile:
            new_profile["rules"] = copy_profile["rules"]


def _create_fixed_profile(autoscale_settings, profile_name, start, end, capacity, copy_rules=None, timezone=None):
    if not (start and end):
        raise CLIError('usage error: fixed schedule: --start DATETIME --end DATETIME')
    profile = {
        "name": profile_name,
        "capacity": capacity,
        "rules": [],
        "fixed_date": {
            "start": start + "+00:00",  # use UTC for AAZDateTimeArg timezone
            "end": end + "+00:00",  # use UTC for AAZDateTimeArg timezone
            "time_zone": timezone
        }
    }
    _apply_copy_rules(autoscale_settings, profile, copy_rules)
    autoscale_settings["profiles"].append(profile)


# pylint: disable=unused-argument
def _create_recurring_profile(autoscale_settings, profile_name, start, end, recurrence, capacity,
                              copy_rules=None, timezone=None):
    from azure.cli.command_modules.monitor._autoscale_util import build_autoscale_profile_dict, \
        validate_autoscale_profile_dict
    import dateutil
    from datetime import time

    def _build_recurrence(base, time):
        recur = {
            "frequency": base["frequency"],
            "schedule": {
                "time_zone": base["schedule"]["time_zone"],
                "days": base["schedule"]["days"],
                "hours": [time.hour],
                "minutes": [time.minute]
            }
        }
        return recur

    start_time = dateutil.parser.parse(start).time() if start else time(hour=0, minute=0)
    end_time = dateutil.parser.parse(end).time() if end else time(hour=23, minute=59)

    default_profile, autoscale_profile = build_autoscale_profile_dict(autoscale_settings)
    validate_autoscale_profile_dict(autoscale_profile, start_time, end_time, recurrence)
    start_profile = {
        "name": profile_name,
        "capacity": capacity,
        "rules": [],
        "recurrence": _build_recurrence(recurrence, start_time)
    }

    _apply_copy_rules(autoscale_settings, start_profile, copy_rules)
    end_profile = {
        "name": json.dumps({'name': default_profile["name"], 'for': profile_name}),
        "capacity": default_profile["capacity"],
        "rules": default_profile["rules"],
        "recurrence": _build_recurrence(recurrence, end_time)
    }
    autoscale_settings["profiles"].append(start_profile)
    autoscale_settings["profiles"].append(end_profile)


def get_autoscale_instance(cmd, resource_group_name, autoscale_name):
    rets = AutoScaleShow(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "autoscale_name": autoscale_name
    })

    rets = _convert_to_snake_case(rets)
    autoscale_settings = {
        "resource_group": resource_group_name,
        "autoscale_name": autoscale_name,
        "enabled": rets.get("enabled", False),
        "tags": rets.get("tags", None),
        "target_resource_location": rets.get("target_resource_location", None),
        "target_resource_uri": rets.get("target_resource_uri", None)
    }
    scale_policy = rets.get("predictive_autoscale_policy", None)
    if scale_policy:
        autoscale_settings["scale_look_ahead_time"] = scale_policy.get("scale_look_ahead_time", None)
        autoscale_settings["scale_mode"] = scale_policy.get("scale_mode", None)
    if rets.get("notifications", None):
        autoscale_settings["notifications"] = rets["notifications"]
    if rets.get("profiles", None):
        autoscale_settings["profiles"] = rets["profiles"]
    return autoscale_settings


def autoscale_profile_create(cmd, autoscale_name, resource_group_name, profile_name,
                             count, timezone, start=None, end=None, copy_rules=None, min_count=None,
                             max_count=None, recurrence=None):
    autoscale_settings = get_autoscale_instance(cmd, resource_group_name, autoscale_name)
    capacity = {
        "default": str(count),
        "minimum": str(min_count) if min_count else str(count),
        "maximum": str(max_count) if max_count else str(count)
    }
    if recurrence:
        _create_recurring_profile(autoscale_settings, profile_name, start, end, recurrence, capacity, copy_rules,
                                  timezone)
    else:
        _create_fixed_profile(autoscale_settings, profile_name, start, end, capacity, copy_rules, timezone)
    updated_rets = AutoScaleUpdate(cli_ctx=cmd.cli_ctx)(command_args=autoscale_settings)
    profile = next(x for x in updated_rets["profiles"] if x["name"] == profile_name)
    return profile


def autoscale_profile_list(cmd, autoscale_name, resource_group_name):
    autoscale_settings = AutoScaleShow(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "autoscale_name": autoscale_name})
    return autoscale_settings["profiles"]


def autoscale_profile_list_timezones(cmd, client, offset=None, search_query=None):
    from azure.cli.command_modules.monitor._autoscale_util import AUTOSCALE_TIMEZONES
    timezones = []
    for zone in AUTOSCALE_TIMEZONES:
        if search_query and search_query.lower() not in zone['name'].lower():
            continue
        if offset and offset not in zone['offset']:
            continue
        timezones.append(zone)
    return timezones


def autoscale_profile_show(cmd, autoscale_name, resource_group_name, profile_name):
    autoscale_settings = AutoScaleShow(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "autoscale_name": autoscale_name})
    return _identify_profile_cg(autoscale_settings["profiles"], profile_name)


class AutoScaleProfileDelete(_AutoScaleUpdate):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.profile_name = AAZStrArg(
            options=["--profile-name"],
            help='Name of the autoscale profile.',
            required=True,
            registered=False,
        )
        return args_schema

    def pre_instance_update(self, instance):
        args = self.ctx.args
        profile_name = args.profile_name.to_serialized_data()
        default_profile = get_autoscale_default_profile(instance)

        def _should_retain_profile(profile):
            name = profile.name.to_serialized_data()
            try:
                name = json.loads(profile.name.to_serialized_data())['for']
            except ValueError:
                pass
            return name.lower() != profile_name.lower()

        retained_profiles = [x for x in instance.properties.profiles if _should_retain_profile(x)]
        instance.properties.profiles = retained_profiles

        # if we removed the last "default" of a recurring pair, we need to preserve it
        new_default = get_autoscale_default_profile(instance)
        if not new_default:
            instance.properties.profiles.append(default_profile)

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        if has_value(self.ctx.vars.instance.properties.name):
            self.ctx.vars.instance.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result


def autoscale_profile_delete(cmd, autoscale_name, resource_group_name, profile_name):
    AutoScaleProfileDelete(cli_ctx=cmd.cli_ctx)(command_args={
        "autoscale_name": autoscale_name,
        "resource_group": resource_group_name,
        "profile_name": profile_name,
    })


def get_condition_from_model(condition):
    condition_obj = {
        "metric_name": condition.metric_name,
        "metric_namespace": condition.metric_namespace,
        "metric_resource_location": condition.metric_resource_location,
        "metric_resource_uri": condition.metric_resource_uri,
        "operator": condition.operator.value,
        "statistic": condition.statistic.value,
        "threshold": float(condition.threshold),
        "time_aggregation": condition.time_aggregation.value,
        "time_grain": condition.time_grain,
        "time_window": condition.time_window
    }
    if condition.divide_per_instance is not None:
        condition_obj["divide_per_instance"] = condition.divide_per_instance
    if condition.dimensions:
        dim_objs = []
        for dim in condition.dimensions:
            dim_objs.append({
                "dimension_name": dim.dimension_name,
                "operator": dim.operator,
                "values": dim.values
            })
        condition_obj["dimensions"] = dim_objs
    return condition_obj


def autoscale_rule_create(cmd, autoscale_name, resource_group_name, condition,
                          scale, profile_name=DEFAULT_PROFILE_NAME, cooldown=5, source=None,
                          timegrain="avg 1m"):
    from azure.mgmt.core.tools import parse_resource_id, resource_id
    autoscale_settings = get_autoscale_instance(cmd, resource_group_name, autoscale_name)
    profile = _identify_profile_cg(autoscale_settings["profiles"], profile_name)
    condition.metric_resource_uri = source or autoscale_settings["target_resource_uri"]
    condition.statistic = timegrain.statistic
    condition.time_grain = timegrain.time_grain

    def preprocess_for_spring_cloud_service():
        try:
            result = parse_resource_id(autoscale_settings["target_resource_uri"])
            if result['namespace'].lower() == 'microsoft.appplatform' and result['type'].lower() == 'spring':
                if condition.metric_namespace is None:
                    condition.metric_namespace = "Microsoft.AppPlatform/Spring"
                    logger.warning('Set metricNamespace to Microsoft.AppPlatform/Spring')
                if source is None:
                    condition.metric_resource_uri = resource_id(
                        subscription=result['subscription'],
                        resource_group=result['resource_group'],
                        namespace=result['namespace'],
                        type=result['type'],
                        name=result['name']
                    )
                    logger.warning('Set metricResourceUri to Spring Cloud service')
        except KeyError:
            pass

    preprocess_for_spring_cloud_service()

    condition_obj = get_condition_from_model(condition)

    rule = {
        "metric_trigger": condition_obj,
        "scale_action": {
            "direction": scale.direction,
            "type": scale.type,
            "cooldown": 'PT{}M'.format(cooldown),
            "value": scale.value
        }
    }
    profile["rules"].append(rule)
    updated_rets = AutoScaleUpdate(cli_ctx=cmd.cli_ctx)(command_args=autoscale_settings)
    updated_profile = _identify_profile_cg(updated_rets["profiles"], profile_name)
    # determine if there are unbalanced rules
    scale_out_rule_count = len([x for x in updated_profile["rules"] if x["scaleAction"]["direction"] == "Increase"])
    scale_in_rule_count = len([x for x in updated_profile["rules"] if x["scaleAction"]["direction"] == "Decrease"])
    if scale_out_rule_count and not scale_in_rule_count:
        logger.warning("Profile '%s' has rules to scale out but none to scale in. "
                       "Recommend creating at least 1 scale in rule.", profile_name)
    elif scale_in_rule_count and not scale_out_rule_count:
        logger.warning("Profile '%s' has rules to scale in but none to scale out. "
                       "Recommend creating at least 1 scale out rule.", profile_name)
    return updated_profile["rules"][-1]


def autoscale_rule_list(cmd, autoscale_name, resource_group_name, profile_name=DEFAULT_PROFILE_NAME):
    autoscale_settings = AutoScaleShow(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "autoscale_name": autoscale_name})
    profile = _identify_profile_cg(autoscale_settings["profiles"], profile_name)
    index = 0
    # we artificially add indices to the rules so the user can target them with the remove command
    for rule in profile["rules"]:
        rule["index"] = index
        index += 1
    return profile["rules"]


def autoscale_rule_delete(cmd, autoscale_name, resource_group_name, index, profile_name=DEFAULT_PROFILE_NAME):
    autoscale_settings = get_autoscale_instance(cmd, resource_group_name, autoscale_name)
    profile = _identify_profile_cg(autoscale_settings["profiles"], profile_name)
    if '*' in index:
        profile["rules"] = []
    else:
        remained_rule = []
        for i, rule in enumerate(profile["rules"]):
            if str(i) in index:
                continue
            remained_rule.append(rule)
        profile["rules"] = remained_rule
    AutoScaleUpdate(cli_ctx=cmd.cli_ctx)(command_args=autoscale_settings)


class AutoScaleRuleCopy(_AutoScaleUpdate):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.source_profile = AAZStrArg(
            options=["--source-profile"],
            help='Name of the autoscale profile.',
            required=True,
            registered=False,
            default=DEFAULT_PROFILE_NAME
        )
        args_schema.dest_profile = AAZStrArg(
            options=["--dest-profile"],
            help='Name of the autoscale profile.',
            required=True,
            registered=False,
        )
        args_schema.index = AAZListArg(
            options=["--index"],
            help="Space-separated list of rule indices to remove, or '*' to clear all rules.",
            registered=False,
        )
        args_schema.index.Element = AAZStrArg()
        return args_schema

    def pre_instance_update(self, instance):
        args = self.ctx.args
        source_profile_name = args.source_profile.to_serialized_data()
        dest_profile_name = args.dest_profile.to_serialized_data()
        index = args.index.to_serialized_data()

        src_profile = _identify_profile(instance.properties.profiles, source_profile_name)
        dst_profile = _identify_profile(instance.properties.profiles, dest_profile_name)
        if '*' in index:
            dst_profile.rules = src_profile.rules
        else:
            for i in index:
                dst_profile.rules.append(src_profile.rules[int(i)])

    def _output(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZUndefined
        if has_value(self.ctx.vars.instance.properties.name):
            self.ctx.vars.instance.properties.name = AAZUndefined
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result


def autoscale_rule_copy(cmd, autoscale_name, resource_group_name, dest_profile, index,
                        source_profile=DEFAULT_PROFILE_NAME):
    AutoScaleRuleCopy(cli_ctx=cmd.cli_ctx)(command_args={
        "autoscale_name": autoscale_name,
        "resource_group": resource_group_name,
        "source_profile": source_profile,
        "dest_profile": dest_profile,
        "index": index
    })


def _identify_profile(profiles, profile_name):
    try:
        profile = next(x for x in profiles if x.name == profile_name)
    except StopIteration:
        raise CLIError('Cannot find profile {}. '
                       'Please double check the name of the autoscale profile.'.format(profile_name))
    return profile


def _identify_profile_cg(profiles, profile_name):
    try:
        profile = next(x for x in profiles if x["name"] == profile_name)
    except StopIteration:
        raise CLIError('Cannot find profile {}. '
                       'Please double check the name of the autoscale profile.'.format(profile_name))
    return profile



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: skip-file
# flake8: noqa
from msrest.serialization import Model


class Contact(Model):
    """The contact information for the vault certificates.

    :param email_address: Email addresss.
    :type email_address: str
    :param name: Name.
    :type name: str
    :param phone: Phone number.
    :type phone: str
    """

    _attribute_map = {
        'email_address': {'key': 'email', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'phone': {'key': 'phone', 'type': 'str'},
    }

    def __init__(self, *, email_address: str=None, name: str=None, phone: str=None, **kwargs) -> None:
        super(Contact, self).__init__(**kwargs)
        self.email_address = email_address
        self.name = name
        self.phone = phone



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: skip-file
# flake8: noqa
from msrest.serialization import Model


class CertificateCreateParameters(Model):
    """The certificate create parameters.

    :param certificate_policy: The management policy for the certificate.
    :type certificate_policy: ~azure.keyvault.v7_0.models.CertificatePolicy
    :param certificate_attributes: The attributes of the certificate
     (optional).
    :type certificate_attributes:
     ~azure.keyvault.v7_0.models.CertificateAttributes
    :param tags: Application specific metadata in the form of key-value pairs.
    :type tags: dict[str, str]
    """

    _attribute_map = {
        'certificate_policy': {'key': 'policy', 'type': 'CertificatePolicy'},
        'certificate_attributes': {'key': 'attributes', 'type': 'CertificateAttributes'},
        'tags': {'key': 'tags', 'type': '{str}'},
    }

    def __init__(self, *, certificate_policy=None, certificate_attributes=None, tags=None, **kwargs) -> None:
        super(CertificateCreateParameters, self).__init__(**kwargs)
        self.certificate_policy = certificate_policy
        self.certificate_attributes = certificate_attributes
        self.tags = tags



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: skip-file
# flake8: noqa
from msrest.serialization import Model


class Action(Model):
    """The action that will be executed.

    :param action_type: The type of the action. Possible values include:
     'EmailContacts', 'AutoRenew'
    :type action_type: str or ~azure.keyvault.v2016_10_01.models.ActionType
    """

    _attribute_map = {
        'action_type': {'key': 'action_type', 'type': 'ActionType'},
    }

    def __init__(self, *, action_type=None, **kwargs) -> None:
        super(Action, self).__init__(**kwargs)
        self.action_type = action_type



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: skip-file
# flake8: noqa
from msrest.serialization import Model


class BackupStorageResult(Model):
    """The backup storage result, containing the backup blob.

    Variables are only populated by the server, and will be ignored when
    sending a request.

    :ivar value: The backup blob containing the backed up storage account.
    :vartype value: bytes
    """

    _validation = {
        'value': {'readonly': True},
    }

    _attribute_map = {
        'value': {'key': 'value', 'type': 'base64'},
    }

    def __init__(self, **kwargs) -> None:
        super(BackupStorageResult, self).__init__(**kwargs)
        self.value = None
